
//----------------------------------------------------------------
//  433.875 with Amazon 3 pin TX module
//----------------------------------------------------------------

#include <ESP8266WiFi.h>
#include <ESP_EEPROM.h>
#include <WiFiUdp.h>
#include <Ticker.h>  

Ticker blinker;
WiFiUDP       UDP;

IPAddress     master_ip(192, 168, 4, 1);
  
IPAddress     remote_ip;
unsigned int  remote_port;
IPAddress     fox_master_ip; 
unsigned int  fox_master_port;

int     TX_PIN = 14;       // D5
int     LED_EXT = 4;       // D2
int     MASTER_SLAVE = 5;  // D1

int     ore = 0;
int     minuti = 0;
int     secondi = 0;

int     fox_next_number = 1;        // fox number assigned to the first slave fox
int     fox_my_number = 0;          // my fox number. master=0, slave1 = 1, etc 
int     fox_minute_next_start_time = 0;

boolean start_tx = false;
boolean synch_requested = false;
boolean DBG = true;
boolean fox_master = false;

struct MyEEPROMStruct 
{
  unsigned char   foxes;
  unsigned char   minute;
} my_eeprom;

// -------------- CONFIGURATION VARIABLES ----------------
boolean use_fixed_parms = false;   // use fixed parms from the program
int     fox_max_number = 2;        // number of foxes in use 
int     fox_start_minute = 1;      // minute at witch we start xmit
//---------------------------------------------------------------------------

//------------------------------
void wifi_setup_master(void)
{
  WiFi.mode(WIFI_AP);
  delay(500);
  WiFi.softAP("ARDF");
  delay(500);  
  

  // start the UDP server
  UDP.begin(1357);

  digitalWrite(LED_EXT, HIGH);
}

//----------------------------------------------------------------------
void wifi_connect(void)
{
  char s[20];
   
  WiFi.begin("ARDF");
 
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
  }

  // start the UDP server
  UDP.begin(1357);

  digitalWrite(LED_EXT, HIGH);
}

//--------------------------------------------------------------------
void ticker_interrupt()
{
  ++secondi;

  if (secondi == 60)
  {
    secondi = 0;
    ++ minuti;
    if (minuti == 60)
    {
      minuti = 0;
      ++ore;
      if (ore == 24)
      {
        ore = 0;
      }
    }
  }

  // if a synch of the time has been requested...
  if (synch_requested == true)
  {
    retrieve_time();
    synch_requested = false;
  }

  // check if we need to start to xmit
  if ((minuti == fox_minute_next_start_time) & (secondi == 0)) 
  {
    start_tx = true;
    // adjust the next start xmit minute
    fox_minute_next_start_time = fox_minute_next_start_time + fox_max_number;
    if (fox_minute_next_start_time > 60)
    {
      fox_minute_next_start_time = fox_minute_next_start_time - 60; 
    }
  }
  
}

//---------------------
void eeprom_load(void)
{
  EEPROM.get(0, my_eeprom);
  fox_max_number = (int) my_eeprom.foxes;
  fox_start_minute = (int) my_eeprom.minute; 
}

//---------------------
void eeprom_save(void)
{
  my_eeprom.foxes = (unsigned char) fox_max_number;
  my_eeprom.minute = (unsigned char) fox_start_minute;
  EEPROM.put(0, my_eeprom);
  EEPROM.commit();  
}


//--------------------------------------------
void setup() 
{
  int  n;
  
  Serial.begin(74880);
  pinMode(LED_EXT, OUTPUT);
  pinMode(TX_PIN, OUTPUT);
  pinMode(MASTER_SLAVE, INPUT);
  digitalWrite(TX_PIN, LOW);
  
  // check if we are a master or slave unit
  n = digitalRead(MASTER_SLAVE);
  if (n == 0)
  {
    fox_master = true;
  } else {
    fox_master = false;
  }
  
  blinker.attach_ms(1000, ticker_interrupt);

  // read the eeprom content and set parms
  EEPROM.begin(sizeof(MyEEPROMStruct));
  
  if (use_fixed_parms == false) eeprom_load();

  // indicate we are up and running 
  for(n=0; n<10; n++)
  {
    digitalWrite(LED_EXT, HIGH);
    delay(200);
    digitalWrite(LED_EXT, LOW);
    delay(200);    
  }

}

//--------------------------------------
void retrieve_send(int a, char s[])
{
  if (a < 10)
  {
    UDP.write('0');
    UDP.write(s[0]);
  } else {
    UDP.write(s[0]);
    UDP.write(s[1]);          
  }
}

//--------------------------
void retrieve_time(void)
{
  char  s[8];

  UDP.beginPacket(remote_ip, remote_port);
  itoa(ore, s, 10);
  retrieve_send(ore, s);
  itoa(minuti, s, 10);
  retrieve_send(minuti, s);
  itoa(secondi, s, 10);
  retrieve_send(secondi, s);
  UDP.endPacket();
}

//---------------------------------------------
void obtain_fox_mumber(void)
{
  char s[4];
   
  UDP.beginPacket(remote_ip, remote_port);
  itoa(fox_next_number, s, 10); 
  UDP.write(s);
  itoa(fox_start_minute, s, 10); 
  UDP.write(s);
  itoa(fox_max_number, s, 10); 
  UDP.write(s);
  
  UDP.endPacket();
    
  ++fox_next_number;
 
}

//------------------------------------
void obtain_phone_data(void)
{
  char  s[4];
  
  // retrieve time
  UDP.beginPacket(remote_ip, remote_port);
 
  itoa(ore, s, 10);
  retrieve_send(ore, s);
  itoa(minuti, s, 10);
  retrieve_send(minuti, s);
  itoa(secondi, s, 10);
  retrieve_send(secondi, s);

  // start minutes and number of foxes around
  itoa(fox_start_minute, s, 10); 
  UDP.write(s);
  itoa(fox_max_number, s, 10); 
  UDP.write(s);
  
  UDP.endPacket(); 
}

//-------------------------------------------
void set_phone_data(char packet[])
{
  int   n, m;
  char  s[4];


  strncpy(s, packet+1, 1);
  n = atoi(s);
  strncpy(s, packet+2, 1);
  m = atoi(s);

  fox_start_minute = n;
  fox_max_number = m;
}

//-------------------------------------------
void check_client_request(void)
{
  char packet[255];

  UDP.read(packet, 255);
  switch(packet[0])
  {
    case 'T' : retrieve_time();
               break;
    case 'S' : synch_requested = true;
               break;           
    case 'G' : obtain_fox_mumber();
               break;     
    case 'P' : obtain_phone_data();
               break;   
    case 'Q' : set_phone_data(packet);
               break;      
  }

  
}

//------------------
void tx_dit(int l)
{
  int  n;

  for(n=0; n<l; n++)
  {
    digitalWrite(TX_PIN, HIGH);
    delayMicroseconds(1000);
    digitalWrite(TX_PIN, LOW);
    delayMicroseconds(1000);
  }
  
}

//--------------------------------------------
void set_cw_caller_id(int id, char s[], int len)
{
  memset(s, 0x00, len);   
  switch(id)
  {
    case 0: strcpy(s, "33033301");      // MOE
            break;  
    case 1: strcpy(s, "330333011");     // MOI
            break;          
    case 2: strcpy(s, "3303330111");    // MOS
            break;         
    case 3: strcpy(s, "33033301111");   // MOH
            break;             
    case 4: strcpy(s, "330333011111");  // MO5
            break;          
  }  
}

//------------------------------------------
void send_cw_id(char s[])
{
  int  n, l;
  int  point_len = 100;      // milliseconds
  
  l = strlen(s);
  for(n=0; n<l; n++)
  {
    switch(s[n])
    {
       case '0' : digitalWrite(LED_EXT, LOW);    
                  delay(point_len * 2);
                  break;
       case '1' : digitalWrite(LED_EXT, HIGH);
                  //delay(point_len);
                  tx_dit(point_len);
                  break;    
       case '3' : digitalWrite(LED_EXT, HIGH);
                  //delay(point_len * 3);
                  tx_dit(point_len * 3);
                  break;                        
    }
    digitalWrite(TX_PIN, LOW);
    digitalWrite(LED_EXT, LOW);
    delay(point_len);
  }

  // end of the word
  digitalWrite(TX_PIN, LOW);
  digitalWrite(LED_EXT, LOW);
  delay(point_len * 9);
}

//---------------------------------------------
void fox_master_code(void)
{
  int   packet_size;

  // if we have just a fox, return, don't need to setup the slave units..
  if (fox_max_number == 1) return;
  
  // setup the other foxes...
  while(true)
  {
    packet_size = UDP.parsePacket();
    if (packet_size > 0)
    {
      remote_ip = UDP.remoteIP();
      remote_port = UDP.remotePort();
      check_client_request();      
    }
    if (fox_next_number == fox_max_number) break;
  }

  // if the configuration has been changed via the APP, update the eeprom content
  if ((fox_max_number != my_eeprom.foxes) | (fox_start_minute != my_eeprom.minute))
  {
    if (use_fixed_parms == false) eeprom_save();
  }
    
}

//-----------------------------------------
void udp_wait_response(char packet[], int l)
{
  int  packet_size;
    
  memset(packet, 0x00, l);
  
  while(true)
  {
    packet_size = UDP.parsePacket();
    if (packet_size > 0)
    {
       UDP.read(packet, l);
       break;
    }
  }  
  
}

//----------------------------------------
void fox_unit_code(void)
{
  char  packet[20];
  char  s[4];
  
  // send synch clock request
  UDP.beginPacket(fox_master_ip, 1357);
  UDP.write("S");
  UDP.endPacket();
  udp_wait_response(packet, 20);

  memset(s, 0x00, sizeof(s));
  strncpy(s, packet+0, 2);
  ore = atoi(s);
  strncpy(s, packet+2, 2);
  minuti = atoi(s);
  strncpy(s, packet+4, 2);
  secondi = atoi(s);
  
  // send get fox number request
  UDP.beginPacket(fox_master_ip, 1357);
  UDP.write("G");
  UDP.endPacket();
  udp_wait_response(packet, 20);

  memset(s, 0x00, sizeof(s));
  strncpy(s, packet+0, 1);
  fox_my_number = atoi(s);
  
  memset(s, 0x00, sizeof(s));
  strncpy(s, packet+1, 1);
  fox_start_minute = atoi(s);
  
  memset(s, 0x00, sizeof(s));
  strncpy(s, packet+2, 1);
  fox_max_number = atoi(s);
  
}

//------------------------------------------
void loop() 
{
  char  s[40];
  int   n;
  int   cnt = 0;
  char  dit_string[15];

  // check if we have to run the master code
  if (fox_master == true)
  {
    wifi_setup_master();
    fox_master_code(); 
  } else {
    wifi_connect();
    fox_unit_code();
  }

  set_cw_caller_id(fox_my_number, dit_string, strlen(dit_string));

  // turn off the radio, we don't need it anymore
  WiFi.mode(WIFI_OFF);
  WiFi.forceSleepBegin();
 
  // compute next minutes at which we'll start to xmit
  fox_minute_next_start_time = fox_start_minute + fox_my_number;
  if (fox_minute_next_start_time > 60)
  {
    fox_minute_next_start_time = fox_minute_next_start_time - 60; 
  }

  digitalWrite(LED_EXT, LOW);
   
  // check if we have to start to xmit
  while(true)
  {
    if (start_tx == true) 
    {
      for(n=0; n<9; n++)
      {
        send_cw_id(dit_string);
      }
      start_tx = false;
    }
   
    delay(10);    
  }

}

